package gbaMus;
import java.io.File;
import java.io.RandomAccessFile;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;

/**
 * GBAMusRiper (c) 2012 by Bregalad
 * This is free and open source software
 * 
 * This program analyze a Game Boy Advance ROM and search for a sound engine
 * named "Sapy" which is used in ~90% of commercial GBA games.
 * 
 * If the engine is found it rips all musics to MIDI (.mid) format and all
 * instruments to SoundFont 2.0 (.sf2) format.
 */

public class GBAMusRiper {
	static RandomAccessFile inGBA;
	static String inGBAPath;
	static String name;
	static String path;
	static boolean gm = false;
	static boolean xg = false;
	static boolean rc = false;
	static boolean sb = false;
	static boolean raw = false;

	final static int[] SampleRates = new int[]{-1, 5734, 7884, 10512, 13379, 15768, 18157, 21024, 26758, 31536, 36314, 40137, 42048};

	public static void main(String[] args) throws IOException {
		//Parse arguments
		parseArgs(args);

		//Auto-detect address of sappy engine
		int sound_engine_adr = SappyDetector.sappyDetector(inGBA);

		//Exit if no sappy engine was found
		if(sound_engine_adr < 0) System.exit(-1);

		//Create a directory named like the input ROM, without the .gba extention
		File dir = new File(name);
		dir.mkdir();

		inGBA.seek(sound_engine_adr + 24);
		//Engine parameter's word
		final int parameter_word = inGBA.readInt();
		//Get sampling rate
		final int sample_rate = SampleRates[(parameter_word &0xf00)>>8];
		final int main_volume = (parameter_word & 0xf00000) >> 20;

		//Compute address of song table
		int song_levels = Integer.reverseBytes(inGBA.readInt());	//Read # of song levels
		int songTblPtr = getPointer(inGBA.readInt()) + 12 * song_levels;

		System.out.println("Parsing song table...");
		//New list of songs
		ArrayList<Integer> song_list = new ArrayList<Integer>();
		//New list of sound banks
		ArrayList<Integer> sound_bank_list = new ArrayList<Integer>();

		inGBA.seek(songTblPtr);

		//Ignores entries which are made of 0s at the start of the song table
		//this fix was necessarily for the game Fire Emblem
		while(inGBA.readInt() == 0)
			songTblPtr += 4;

		inGBA.seek(songTblPtr);
		int song_pointer;
		int i = 0;
		while(true){
			song_pointer = getPointer(inGBA.readInt());		//Read song pointer

			//Stop as soon as we met with an invalid pointer
			if(song_pointer <= 0 || song_pointer >= inGBA.length())
				break;

			inGBA.readInt();							//Ignore sound group
			song_list.add(song_pointer);				//Add pointer to list

			i++;
		};
		//As soon as data that is not a valid pointer is found, the song table is terminated

		//End of song table
		int songTblEndPtr = 8 * i + songTblPtr;

		System.out.println("Collecting sound bank list...");

		int[] sound_bank_index_list = new int[song_list.size()];
		for(i = 0; i < song_list.size(); i++) {
			//Ignore unused song, which points to the end of the song table (for some reason)
			if(song_list.get(i) != songTblEndPtr) {
				//Seek to song data
				inGBA.seek(song_list.get(i) + 4);
				int soundBankPtr = getPointer(inGBA.readInt());
	
				//Add sound bank to list if not already in the list
				if(!sound_bank_list.contains(soundBankPtr))
					sound_bank_list.add(soundBankPtr);
	
				sound_bank_index_list[i] = sound_bank_list.indexOf(soundBankPtr);
			}
		}

		//Close GBA file so that SongRiper can access it
		inGBA.close();
		
		//Create directories for each sound bank if separate banks is enabled
		if(sb) {
			for(i=0; i < sound_bank_list.size(); i++) {
				String number = String.format(String.format("%%0%dd", 3), i);
				File subdir = new File(name + File.separatorChar + "soundbank_" + number);
				subdir.mkdir();
			}
		}

		for(i = 0; i < song_list.size(); i++) {
			if(song_list.get(i) != songTblEndPtr) {
				String[] seqRipArgs = new String[raw ? (sb ? 4 : 5) : (sb ? 6 : 7)];
	
				seqRipArgs[0] = inGBAPath;

				//Add leading zeroes to file name
				String number = String.format(String.format("%%0%dd", 3), i);
				String sbnumber = String.format(String.format("%%0%dd", 3), sound_bank_index_list[i]);

				if(sb)
					seqRipArgs[1] = name + File.separatorChar + "soundbank_" + sbnumber +
									File.separatorChar + "song" + number + ".mid";
				else
					seqRipArgs[1] = name + File.separatorChar + "song" + number + ".mid";

				seqRipArgs[2] = "0x" + Integer.toHexString(song_list.get(i));
				seqRipArgs[3] = rc ? "-rc" : (xg ? "-xg" : "-gs");
				if(!raw) {
					seqRipArgs[4] = "-sv";
					seqRipArgs[5] = "-lv";
				}
				//Bank number, if banks are not separated
				if(!sb)
					seqRipArgs[raw ? 4 : 6] = "-b" + sound_bank_index_list[i];
	
				System.out.println("Song " + i);
				try{
					SongRiper.main(seqRipArgs);
				}catch(Exception e) {System.out.println("An exception occured");}
			}
		}

		if(sb) {
			//Rips each sound bank in a different file/folder
			for(i = 0; i < sound_bank_list.size(); i++) {
				String[] sfRipArgs = new String[gm ? 6 : 5];

				sfRipArgs[0] = inGBAPath;

				String sbnumber = String.format(String.format("%%0%dd", 3), i);
				String sbname = "soundbank_" + sbnumber + "_@" + Integer.toHexString(sound_bank_list.get(i));
				sfRipArgs[1] = name + File.separatorChar + "soundbank_" + sbnumber + File.separatorChar + name + "_" + sbname + ".sf2";
				
				sfRipArgs[2] = "-s" + Integer.toString(sample_rate);
				sfRipArgs[3] = "-mv" + Integer.toString(main_volume);
				if(gm)
					sfRipArgs[4] = "-gm";
				sfRipArgs[gm? 5 : 4] = "0x" + Integer.toHexString(sound_bank_list.get(i));
				
				SoundFontRiper.main(sfRipArgs);
			}
		}
		else {
			//Rips each sound bank in a single soundfont file
			//Build argument list to call SoundFontRipper
			String[] sfRipArgs = new String[sound_bank_list.size() + (gm ? 5 : 4)];
			sfRipArgs[0] = inGBAPath;
			//Output sound font named after the input ROM
			sfRipArgs[1] = name + File.separatorChar + name + ".sf2";
			sfRipArgs[2] = "-s" + Integer.toString(sample_rate);
			sfRipArgs[3] = "-mv" + Integer.toString(main_volume);
			//Pass -gm argument if necessary
			if(gm)
				sfRipArgs[4] = "-gm";
	
			//Make sound banks addresses list.
			for(i=0; i < sound_bank_list.size(); i++)
				sfRipArgs[i+(gm ? 5 : 4)] = "0x" + Integer.toHexString(sound_bank_list.get(i));
	
			//Call sound font riper
			SoundFontRiper.main(sfRipArgs);
		}
		System.out.println("Rip completed !");
	}

	public static int getPointer(int Pointer) {
		//Fix pointer to ROM pointer
		return Integer.reverseBytes(Pointer) - 0x8000000;
	}

	static void parseArgs(String[] args) throws IOException {
		if(args.length < 1)
			printInstructions();

		inGBAPath = args[0];
		//Get GBA file
		inGBA = new RandomAccessFile(inGBAPath, "r");

		//Name is filename without the extention and without path
		name = args[0].substring(args[0].lastIndexOf(File.separatorChar) + 1, args[0].lastIndexOf("."));

		//Path where the input GBA file is located
		path = args[0].substring(0, args[0].lastIndexOf(File.separatorChar) + 1);

		for(int i = 1; i<args.length; i++) {
			if(args[i].equals("-gm"))
				gm = true;
			else if(args[i].equals("-xg"))
				xg = true;
			else if(args[i].equals("-rc"))
				rc = true;
			else if(args[i].equals("-sb"))
				sb = true;
			else if(args[i].equals("-raw"))
				raw = true;
			else
				printInstructions();
		}
	}

	static void printInstructions() {
		System.out.println("GBAMusRiper (c) 2012 by Bregalad");
		System.out.println("Usage : GBAMusRiper game.gba [flags]");
		System.out.println("-gm : Give General MIDI names to presets. Note that this will only change the names and will NOT " +
							"magically turn the soundfont into a General MIDI compliant soundfont.");
		System.out.println("-xg : Output MIDI will be compliant to XG standard (instead of default GS standard)");
		System.out.println("-rc : Rearrange channels in output MIDIs so channel 10 is avoided. Needed by sound" +
				"cards where it's impossible to disable \"drums\" on channel 10 even with GS or XG commands");
		System.out.println("-sb : Separate banks. Every sound bank is riper to a different .sf2 file and placed" +
				" into different sub-folders (instead of doing it in a single .sf2 file and a single folder)");
		System.out.println("-raw : Output MIDIs exactly as they're encoded in ROM, without linearize volume and" +
				" velocities and without simulating vibratoes.");
		System.exit(-1);
	}

}
