package gbaMus;

import java.io.DataOutputStream;
import java.io.IOException;

/**
 * Sound Font main class
 * 
 * This is part of the GBA SoundFont Ripper (c) 2012 by Bregalad
 * This is free and open source software.
 * 
 * Notes : I tried to separate the GBA-related stuff and SF2 related stuff as much as possible in different classes
 * that way anyone can re-use this program for building SF2s out of different data.
 * 
 * SF2 is a SoundFont file format. This class (and related classes) serves to create data to build
 * a SF2 file. For more details look at the doccument "Sound Font(R) Technical Doccumentation" version 2.01
 * which was used as a reference when writing these classes.
 * By the way I must tell this format is incredibly stupid and complex as opposed to what it could be and
 * I have no idea why people started to adopt this standard. But anyways they did...
 * 
 * All SF2-related classes contains a field named sf2 that links to the SF2 main class to know which
 * SF2 file they relates to. (this would make it possible to build multiple SF2 files at a time).
 * 
 * The building of a SF2 file is done in multiple passes :
 * - Creating samples
 * - Creating instruments
 * - Creating presets
 * - Output to the target file
 */

public class SF2 {

	int DefaultSampleRate = 22050;
	int size = 0;

	//Instrument and samples objects
	SF2Instr instruments;
	SF2Samples samples;

	//All 3 chunks of the SF2 file
	InfoListChunk infoListChunk;
	SdtaListChunk sdtaListChunk;
	HydraChunk pdtaListChunk;

	//Target file
	DataOutputStream out;

	//Create a blank SF2 structure
	SF2() {
		instruments = new SF2Instr(this);
		samples = new SF2Samples(this);
		infoListChunk = new InfoListChunk(this);
		sdtaListChunk = new SdtaListChunk(this);
		pdtaListChunk = new HydraChunk(this);
	}

	//Constructor to change the default sample rate
	SF2(int sampleRate) {
		this();
		DefaultSampleRate = sampleRate;
	}

	//Write data to the target file
	//(should only be called once !)
	void write(DataOutputStream outfile) throws IOException {
		out = outfile;
		
		//This function adds the "terminal" data in subchunks that are required
		//by the (retarded) SF2 standard
		addTerminals();

		//Compute size of the entire file
		//(this will also compute the size of the chunks)
		size = 4;
		size += infoListChunk.calcSize() + 8;
		size += sdtaListChunk.calcSize() + 8;
		size += pdtaListChunk.calcSize() + 8;

		//Write RIFF header
		out.writeBytes("RIFF");
		out.writeInt(Integer.reverseBytes(size));
		out.writeBytes("sfbk");

		//Write all 3 chunks
		infoListChunk.write();
		sdtaListChunk.write();
		pdtaListChunk.write();

		//Close output file
		out.close();
	}

	//Function to write a string of 20 characters,
	//zero terminated if length smaller than 20
	void writeString20(String s) throws IOException {
		//Truncate the string if longer than 20 characters
		if(s.length()>20)
			s = s.substring(0, 20);

		out.writeBytes(s);
		for(int i = 20; i != s.length(); i--)
			out.write(0x00);
	}

	//Add terminal data in subchunks where it is required by the standard
	void addTerminals() {
		add_new_sample_header("EOS", 0, 0, 0, 0, 0, 0, 0);
		add_new_instrument("EOI");
		add_new_inst_bag();
		add_new_inst_generator();
		add_new_inst_modulator();
		add_new_preset("EOP", 255, 255);
		add_new_preset_bag();
		add_new_preset_generator();
		add_new_preset_modulator();
	}

	//Add a brand new preset header to the list
	void add_new_preset(String name, int Patch, int Bank) {
		sfPresetHeader new_preset = new sfPresetHeader(this, name, Patch, Bank);
		pdtaListChunk.phdr_subchunk.add_preset(new_preset);
	}

	//Add a brand new instrument
	void add_new_instrument(String name) {
		sfInst new_inst = new sfInst(this, name);
		pdtaListChunk.inst_subchunk.add_instrument(new_inst);
	}

	//Add a new instrument bag to the instrument bag list
	//DO NOT use this to add a preset bag !
	void add_new_inst_bag() {
		sfBag newbag = new sfBag(this, false);
		pdtaListChunk.ibag_subchunk.add_bag(newbag);
	}

	//Add a new preset bag to the preset bag list
	//DO NOT use this to add an instrument bag !
	void add_new_preset_bag() {
		sfBag newbag = new sfBag(this, true);
		pdtaListChunk.pbag_subchunk.add_bag(newbag);
	}

	//Add a new modulator to the list
	void add_new_preset_modulator() {
		sfModList newmod = new sfModList(this);
		pdtaListChunk.pmod_subchunk.add_modulator(newmod);
	}

	//Add a new blank generator to the list
	void add_new_preset_generator() {
		sfGenList new_generator = new sfGenList(this);
		pdtaListChunk.pgen_subchunk.add_generator(new_generator);
	}

	//Add a new customized generator to the list
	void add_new_preset_generator(SFGenerator operation, int value) {
		sfGenList new_generator = new sfGenList(this, operation, new genAmountType((short)value));
		pdtaListChunk.pgen_subchunk.add_generator(new_generator);
	}

	//Add a new customized generator to the list
	void add_new_preset_generator(SFGenerator operation, int lo, int hi) {
		sfGenList new_generator = new sfGenList(this, operation, new genAmountType((byte)lo, (byte)hi));
		pdtaListChunk.pgen_subchunk.add_generator(new_generator);
	}

	//Add a new modulator to the list
	void add_new_inst_modulator() {
		sfModList newmod = new sfModList(this);
		pdtaListChunk.imod_subchunk.add_modulator(newmod);
	}

	//Add a new blank generator to the list
	void add_new_inst_generator() {
		sfGenList new_generator = new sfGenList(this);
		pdtaListChunk.igen_subchunk.add_generator(new_generator);
	}

	//Add a new customized generator to the list
	void add_new_inst_generator(SFGenerator operation, int value) {
		sfGenList new_generator = new sfGenList(this, operation, new genAmountType((short)value));
		pdtaListChunk.igen_subchunk.add_generator(new_generator);
	}

	//Add a new customized generator to the list
	void add_new_inst_generator(SFGenerator operation, int lo, int hi) {
		sfGenList new_generator = new sfGenList(this, operation, new genAmountType((byte)lo, (byte)hi));
		pdtaListChunk.igen_subchunk.add_generator(new_generator);
	}

	//Add a brand new header
	void add_new_sample_header(String name, int Start, int End, int StartLoop,
								int EndLoop, int SampleRate,int OriginalPitch, int PitchCorrection) {
		sfSample newspl = new sfSample(this, name, Start, End, StartLoop, EndLoop,
										SampleRate, OriginalPitch, PitchCorrection);
		pdtaListChunk.shdr_subchunk.add_sample(newspl);
	}

}
