package gbaMus;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;

/**
 * Sound Font chunk classes
 * 
 * This is part of the GBA SoundFont Riper (c) 2012 by Bregalad
 * This is free and open source software.
 * 
 * Notes : I tried to separate the GBA-related stuff and SF2 related stuff as much as possible in different classes
 * that way anyone can re-use this program for building SF2s out of different data.
 * 
 * SF2 is a SoundFont file format. This class (and related classes) serves to create data to build
 * a SF2 file. For more details look at the document "Sound Font(R) Technical Documentation" version 2.01
 * which was used as a reference when writing these classes.
 * By the way I must tell this format is incredibly stupid and complex as opposed to what it could be and
 * I have no idea why people started to adopt this standard. But anyways they did...
 * 
 * This file contains classes for all chunks and subchunks present in a SF2 file
 * 
 * All SF2-related classes contains a field named sf2 that links to the SF2 main class to know which
 * SF2 file they relates to. (this would make it possible to build multiple SF2 files at a time).
 * 
 * Note : These classes helps in the complex process of building a SF2 and make it much more simpler and automated.
 * Not all of SF2's features are supported, support for ROM samples, stereo samples and modulators is
 * incomplete Yet those classes contains nothing GBA-related and could easily be re-used in another project as-it.
 * 
 * There is basically 2 ways to adds data, such as a new sample, new instrument of new preset.
 * 
 * The first way is to make it all automatically by using the add_new_xxx() functions. It's recommended to do
 * it this way whenever possible HOWEVER to have a file which is structurally correct it is imperative to
 * add them in this order :
 * 1) Instrument/Preset header
 * 2) Bag 1
 * 3) Modulators and generators for Bag1
 * 4) Bag 2 (optional)
 * 5) Modulators and generators for Bag2 (optional)
 * etc...
 * 
 * The second way is to create moderators, generators and bag directly using their respective classes. That way
 * they can be created in any order, but it's important they are created in a first step, and added in a second step
 * in the order they should be in the final file.
 */

//SF2Chunks abstract class
//All Chunks and SubChunks should extend this class.
//They should all call this constructor first,
//and they should have a write() function that calls this write() function first.
abstract class SF2Chunks {
	int size;		//Size in bytes of the subchunk
	String name;	//4-letter name of the subchunk
	SF2 sf2;

	//Constructor (should be systematically called by sub-classes)
	SF2Chunks(SF2 sf2){
		this.sf2 = sf2;	//Link to output SF2 file
		size = 0;		//The chunk starts bank
	}

	//Write the name and size of the (sub)chunk (should be systematically called by sub-classes)
	void write() throws IOException {
		sf2.out.writeBytes(name);
		sf2.out.writeInt(Integer.reverseBytes(size));
	}
}

//Info List chunk, containing info about the SF2 file
class InfoListChunk extends SF2Chunks {
	//Sub-chunks in InfoList chunk
	IFILSubChunk ifil_subchunk;
	HeaderSubChunk isng_subchunk;
	HeaderSubChunk inam_subchunk ;
	HeaderSubChunk ieng_subchunk;
	HeaderSubChunk icop_subchunk;

	//Info List constructor
	InfoListChunk(SF2 sf2) {
		super(sf2);
		ifil_subchunk = new IFILSubChunk(sf2);
		isng_subchunk = new HeaderSubChunk(sf2, "isng", "EMU8000");
		inam_subchunk = new HeaderSubChunk(sf2, "INAM", "Unnamed");
		ieng_subchunk = new HeaderSubChunk(sf2, "IENG", "Nintendo Game Boy Advance SoundFont");
		icop_subchunk = new HeaderSubChunk(sf2, "ICOP", "Ripped with SF2Ripper v0.0 (c) 2012 by Bregalad");

		name = "LIST";
	}

	//Compute size of the info-list chunk
	int calcSize() {
		size = 4;
		size += ifil_subchunk.size + 8;
		size += isng_subchunk.size + 8;
		size += inam_subchunk.size + 8;
		size += ieng_subchunk.size + 8;
		size += icop_subchunk.size + 8;

		return size;
	}

	void write() throws IOException {
		super.write();				//Write chunk name and size
		sf2.out.writeBytes("INFO");	//Chunk header

		ifil_subchunk.write();		//Write all 5 sub-chunk elements
		isng_subchunk.write();
		inam_subchunk.write();
		ieng_subchunk.write();
		icop_subchunk.write();
	}
}

//Sample data list chunk, contains samples
class SdtaListChunk extends SF2Chunks {
	SMPLSubChunk smpl_subchunk;

	SdtaListChunk (SF2 sf2) {
		super(sf2);
		smpl_subchunk = new SMPLSubChunk(sf2);
		name = "LIST";
	}

	//Compute size of sample-data-list chunk
	int calcSize() {
		size = 4;
		size += smpl_subchunk.size + 8;

		return size;
	}

	void write() throws IOException {
		super.write();
		sf2.out.writeBytes("sdta");

		smpl_subchunk.write();
	}
}

//Hydra chunk, contains data for instruments, presets and samples header
class HydraChunk extends SF2Chunks {
	//Sub chunks in PTDA chunk
	PHDRSubChunk phdr_subchunk;
	BAGSubChunk pbag_subchunk;
	MODSubChunk pmod_subchunk;
	GENSubChunk pgen_subchunk;
	INSTSubChunk inst_subchunk;
	BAGSubChunk ibag_subchunk;
	MODSubChunk imod_subchunk;
	GENSubChunk igen_subchunk;
	SHDRSubChunk shdr_subchunk;

	//Constructor
	HydraChunk(SF2 sf2) {
		super(sf2);
		name = "LIST";

		phdr_subchunk = new PHDRSubChunk(sf2);
		pbag_subchunk = new BAGSubChunk(sf2, true);
		pmod_subchunk = new MODSubChunk(sf2, true);
		pgen_subchunk = new GENSubChunk(sf2, true);
		inst_subchunk = new INSTSubChunk(sf2);
		ibag_subchunk = new BAGSubChunk(sf2, false);
		imod_subchunk = new MODSubChunk(sf2, false);
		igen_subchunk = new GENSubChunk(sf2, false);
		shdr_subchunk = new SHDRSubChunk(sf2);
	}

	int calcSize() {
		size = 4;
		//Compute size of the chunk (because we don't know it in advance)
		size += phdr_subchunk.size + 8;
		size += pbag_subchunk.size + 8;
		size += pmod_subchunk.size + 8;
		size += pgen_subchunk.size + 8;
		size += inst_subchunk.size + 8;
		size += ibag_subchunk.size + 8;
		size += imod_subchunk.size + 8;
		size += igen_subchunk.size + 8;
		size += shdr_subchunk.size + 8;

		return size;
	}

	void write() throws IOException {
		//Write chunk name and size
		super.write();
		sf2.out.writeBytes("pdta");

		//Write all sub-chunks
		phdr_subchunk.write();
		pbag_subchunk.write();
		pmod_subchunk.write();
		pgen_subchunk.write();
		inst_subchunk.write();
		ibag_subchunk.write();
		imod_subchunk.write();
		igen_subchunk.write();
		shdr_subchunk.write();
	}
}

//Version sub-chunk
class IFILSubChunk extends SF2Chunks {
	//Output format is SoundFont v2.1
	final short wMajor = 2;
	final short wMinor = 1;

	IFILSubChunk(SF2 sf2) {
		super(sf2);
		name = "ifil";
		size = 4;
	}

	void write() throws IOException {
		super.write();
		sf2.out.writeShort(Short.reverseBytes(wMajor));
		sf2.out.writeShort(Short.reverseBytes(wMinor));
	}
}

//Class for the various header chunks that just contain a string
class HeaderSubChunk extends SF2Chunks {
	final String field;

	HeaderSubChunk(SF2 sf2, String subchunk_type, String s) {
		super(sf2);
		name = subchunk_type;
		field = s;
		size = s.length() + 1;	//The string is null terminated -> it takes one more byte
	}

	void write() throws IOException {
		super.write();
		sf2.out.writeBytes(field);	//Write the string followed by a null byte
		sf2.out.writeByte(0x00);
	}
}

enum SampleType {
	unsigned_8,
	signed_8,
	signed_16,
	gameboy_ch3;
}

//Class for the samples sub chunk
class SMPLSubChunk extends SF2Chunks {
//	ArrayList<short[]> sample_list;		//List of samples
	
	//To prevent the program from using a lot of memory by caching all
	//samples before writing them (which is not useful)
	//I instead store a list of pointers to sample data, and the data
	//is directly read from the original file when the sample should
	//be written to output

	//I made this function as generic as possible, so any sample can be loaded
	//from any file, in various formats.

	ArrayList<RandomAccessFile> file_list;	//Files from which the samples must be read
	ArrayList<Integer> pointer_list;		//address within files where the samples must be read
	ArrayList<Integer> size_list;			//Size of the data sample
	ArrayList<Boolean> loop_flag_list;		//Loop flag for samples (required as we need to copy data after the loop)
	ArrayList<Integer> loop_pos_list;		//Loop start data (irrelevent if loop flag is clear - add dummy data)
	ArrayList<SampleType> sample_type_list;	//Type of sample (unsigned / signed, 8/16 bits etc...)

	SMPLSubChunk(SF2 sf2) {
		super(sf2);
		name = "smpl";
		file_list = new ArrayList<RandomAccessFile>();
		pointer_list = new ArrayList<Integer>();
		size_list = new ArrayList<Integer>();
		loop_flag_list = new ArrayList<Boolean>();
		loop_pos_list = new ArrayList<Integer>();
		sample_type_list = new ArrayList<SampleType>();
	}

	//Add a sample to the package
	//Returns directory index of the start of the sample
	int addSample(RandomAccessFile file, SampleType type, int pointer, int size, boolean loop_flag, int loop_pos) {
		file_list.add(file);
		pointer_list.add(pointer);
		size_list.add(size);
		loop_flag_list.add(loop_flag);
		loop_pos_list.add(loop_pos);
		sample_type_list.add(type);

		int dir_offset = this.size >> 1;
		//2 bytes per sample
		//Compute size including the 8 samples after loop point
		//and 46 dummy samples
		if(loop_flag)
			this.size += (size + 8 + 46) * 2;
		else
			this.size += (size + 46) * 2;

		return dir_offset;
	}

	//Add a new sample and create corresponding header
	void addNewSample(RandomAccessFile file, SampleType type, String name, int pointer, int size,
						boolean loop_flag, int loop_pos, int OriginalPitch, int PitchCorrection, int SampleRate) {

		int dir_offset = addSample(file, type, pointer, size, loop_flag, loop_pos);
		// If the sample is looped SF2 standard requires we add the 8 bytes
		// at the start of the loop at the end (what a dumb standard)
		int dir_end, dir_loop_end, dir_loop_start;

		if (loop_flag) {
			dir_end = dir_offset + size + 8;
			dir_loop_end = dir_offset + size;
			dir_loop_start = dir_offset + loop_pos;
		} else {
			dir_end = dir_offset + size;
			dir_loop_end = 0;
			dir_loop_start = 0;
		}

		//Create sample header and add it to the list
		sf2.add_new_sample_header(name, dir_offset, dir_end, dir_loop_start, dir_loop_end,
							SampleRate, OriginalPitch, PitchCorrection);
	}

	//Add new sample using default sample rate
	void addNewSample(RandomAccessFile file, SampleType type, String name, int pointer, int size,
						boolean loop_flag, int loop_pos, int OriginalPitch, int PitchCorrection) {

		addNewSample(file, type, name, pointer, size, loop_flag, loop_pos,
						OriginalPitch, PitchCorrection, sf2.DefaultSampleRate);
	}

	//Write all samples to output in little Indian format
	void write() throws IOException {
		super.write();

		for(int i=0; i<file_list.size(); i++) {
			//Seek at the start of the sample in input file
			file_list.get(i).seek(pointer_list.get(i));

			//Using a cached buffer really speeds up the writing process a lot !!
			byte[] outbuf = new byte[2*size_list.get(i)];

			switch (sample_type_list.get(i)) {
			
			//Source is unsigned 8 bits
			case unsigned_8: {
				byte[] data = new byte[size_list.get(i)];
				file_list.get(i).read(data);

				for(int j=0; j < size_list.get(i); j++) {
					outbuf[2*j] = 0x00;
					outbuf[2*j+1] = (byte)(data[j] - 0x80);
				}
			}break;

			//Source is signed 8 bits
			case signed_8: {
				byte[] data = new byte[size_list.get(i)];
				file_list.get(i).read(data);

				for(int j=0; j < size_list.get(i); j++) {
					outbuf[2*j] =  0x00;
					outbuf[2*j+1] = data[j];
				}
			}break;
			
			case signed_16: {
				//Copy the sample as-is
				file_list.get(i).read(outbuf);
			}break;
			
			case gameboy_ch3: {
				//Conversion lookup table
				final byte conv_tbl[] = {-80, -70, -60, -50, -40, -30, -20, -1, 0, 10, 20, 30, 40, 50, 60, 70};

				int num_of_repts = size_list.get(i)/32;
				//Data is always on 16 bytes
				byte[] data = new byte[16];
				file_list.get(i).read(data);

				for(int j=0, l=0; j<16; j++) {
					for(int k=0; k<num_of_repts; k++, l++)
						outbuf[2*l+1] = conv_tbl[(data[j]>>4)&0xf];
					for(int k=0; k<num_of_repts; k++, l++)
						outbuf[2*l+1] = conv_tbl[data[j]&0xf];
				}
			}break;
			}

			//Write buffer
			sf2.out.write(outbuf);

			//If loop enabled, write 8 samples after loop point
			//(required by the dumb SF2 standard)
			if(loop_flag_list.get(i))
				sf2.out.write(outbuf, 2*loop_pos_list.get(i), 16);

			//Write 46 dummy zeroed samples at the end
			//which is also required by the very dumb SF2 standard
/*			for(int j = 0; j < 46; j++)
				sf2.out.writeShort(0x00);	*/
			byte[] dummy = new byte[2*46];
			sf2.out.write(dummy);
		}
	}
}

//Preset header list sub-chunk
class PHDRSubChunk extends SF2Chunks {
	ArrayList<sfPresetHeader> preset_list;

	PHDRSubChunk(SF2 sf2) {			//Init size and name
		super(sf2);
		name = "phdr";
		preset_list = new ArrayList<sfPresetHeader>();
	}

	//Add an existing preset header to the list
	void add_preset(sfPresetHeader preset) {
		preset_list.add(preset);
		size += 38;				//Each entry is exactly 38 bytes long
	}

	void write() throws IOException {
		super.write();

		//Call the write function for all elements of the list
		for(int i=0; i<preset_list.size(); i++)
			preset_list.get(i).write();
	}
}

//Instrument list sub chunk
class INSTSubChunk extends SF2Chunks {
	ArrayList<sfInst> instrument_list;

	INSTSubChunk (SF2 sf2) {
		super(sf2);
		name = "inst";
		instrument_list = new ArrayList<sfInst>();
	}

	//Add an existing instrument
	void add_instrument(sfInst instrument) {
		instrument_list.add(instrument);
		size += 22;
	}

	void write() throws IOException {
		super.write();

		for(int i=0; i<instrument_list.size(); i++)
			instrument_list.get(i).write();
	}
}

// Preset/Instrument bag list sub chunk
class BAGSubChunk extends SF2Chunks {
	ArrayList<sfBag> bag_list;

	BAGSubChunk (SF2 sf2, boolean preset) {		//Init size and name
		super(sf2);
		if(preset)
			name = "pbag";
		else
			name = "ibag";
		bag_list = new ArrayList<sfBag>();
	}

	//Add an existing bag to the list
	void add_bag(sfBag bag) {
		bag_list.add(bag);
		size += 4;				//Each entry is exactly 4 bytes long
	}

	void write() throws IOException {
		super.write();

		//Call the write function for all elements of the list
		for(int i=0; i<bag_list.size(); i++)
			bag_list.get(i).write();
	}
}

//Preset/Instrument Modulator list class
class MODSubChunk extends SF2Chunks {
	ArrayList<sfModList> modulator_list;

	MODSubChunk (SF2 sf2, boolean preset) {
		super(sf2);
		if(preset)
			name = "pmod";
		else
			name = "imod";
		modulator_list = new ArrayList<sfModList>();
	}

	//Add an existing modulator to the list
	void add_modulator(sfModList modulator) {
		modulator_list.add(modulator);
		size += 10;
	}

	void write() throws IOException {
		super.write();

		for(int i=0; i<modulator_list.size(); i++)
			modulator_list.get(i).write();
	}
}

//Preset/Instrument Generator list class
class GENSubChunk extends SF2Chunks {
	ArrayList<sfGenList> generator_list;

	GENSubChunk (SF2 sf2, boolean preset) {
		super(sf2);
		if(preset)
			name = "pgen";
		else
			name = "igen";

		generator_list = new ArrayList<sfGenList>();
	}

	//Add an existing generator to the list
	void add_generator(sfGenList generator) {
		generator_list.add(generator);
		size += 4;
	}

	void write() throws IOException {
		super.write();

		for(int i=0; i<generator_list.size(); i++)
			generator_list.get(i).write();
	}
}

//Sample header list class
class SHDRSubChunk extends SF2Chunks {
	ArrayList<sfSample> sample_list;

	SHDRSubChunk(SF2 sf2) {
		super(sf2);
		name = "shdr";
		sample_list = new ArrayList<sfSample>();
	}

	//Add an existing header to the list
	void add_sample(sfSample sample) {
		sample_list.add(sample);
		size += 46;
	}

	void write() throws IOException {
		super.write();

		for(int i=0; i<sample_list.size(); i++)
			sample_list.get(i).write();
	}
}

//Preset header class
class sfPresetHeader {
	String achPresetName;			//Preset's name
	short wPreset;					//Patch #
	short wBank;					//Bank #
	private short wPresetBagNdx;	//Index to "bag" of instruments (private - created automatically)
	final int dwLibrary = 0;		//Unused values - should be kept to 0
	final int dwGenre = 0;
	final int dwMorphology = 0;
	SF2 sf2;

	sfPresetHeader(SF2 sf2, String name, int Patch, int Bank) {
		this.sf2 = sf2;
		achPresetName = name;
		wPreset = (short)Patch;
		wBank = (short)Bank;
		wPresetBagNdx = (short)sf2.pdtaListChunk.pbag_subchunk.bag_list.size();
	}

	void write() throws IOException {
		sf2.writeString20(achPresetName);

		sf2.out.writeShort(Short.reverseBytes(wPreset));
		sf2.out.writeShort(Short.reverseBytes(wBank));
		sf2.out.writeShort(Short.reverseBytes(wPresetBagNdx));
	
		sf2.out.writeInt(Integer.reverseBytes(dwLibrary));
		sf2.out.writeInt(Integer.reverseBytes(dwGenre));
		sf2.out.writeInt(Integer.reverseBytes(dwMorphology));
	}
}

//Preset bag class
class sfBag {
	//Private to prevent anyone from affecting the
	//indexes which are automatically created....
	private short wGenNdx;			//Index to list of generators
	private short wModNdx;			//Index to list of modulators
	SF2 sf2;

	//Automatically assign indexes 
	sfBag(SF2 sf2, boolean Preset) {
		this.sf2 = sf2;
		if(Preset) {
			wGenNdx = (short)sf2.pdtaListChunk.pgen_subchunk.generator_list.size();
			wModNdx = (short)sf2.pdtaListChunk.pmod_subchunk.modulator_list.size();
		}
		else {
			wGenNdx = (short)sf2.pdtaListChunk.igen_subchunk.generator_list.size();
			wModNdx = (short)sf2.pdtaListChunk.imod_subchunk.modulator_list.size();
		}
	}

	void write() throws IOException {
		sf2.out.writeShort(Short.reverseBytes(wGenNdx));
		sf2.out.writeShort(Short.reverseBytes(wModNdx));
	}
}

//Modulator class (mostly unused)
class sfModList {
	SFModulator sfModSrcOper;		//Modulator source
	SFGenerator sfModDestOper;		//Modulator destination
	short modAmount;				//Modulator value
	SFModulator sfModAmtSrcOper;	//Modulator source ??
	SFTransform sfModTransOper;		//Transformation curvative
	SF2 sf2;

	sfModList(SF2 sf2) {
		this.sf2 = sf2;
		sfModSrcOper = SFModulator._null;
		sfModDestOper = SFGenerator._null;
		modAmount = 0;
		sfModAmtSrcOper = SFModulator._null;
		sfModTransOper = SFTransform._null;
	}

	void write() throws IOException {
		sf2.out.writeShort(Short.reverseBytes(sfModSrcOper.value));
		sf2.out.writeShort(Short.reverseBytes(sfModDestOper.value));
		sf2.out.writeShort(Short.reverseBytes(modAmount));
		sf2.out.writeShort(Short.reverseBytes(sfModAmtSrcOper.value));
		sf2.out.writeShort(Short.reverseBytes(sfModTransOper.value));
	}
}

//Java equivalent of a C union...
//Two bytes that can handle either two 8-bit values or a single 16-bit value
class genAmountType {
	final short value;

	genAmountType(short a) {
		value = a;
	}

	genAmountType(byte lo, byte hi) {
		value = (short)(hi << 8 | lo);
	}
}

//Generator class
//This is extremely important
class sfGenList {
	SFGenerator sfGenOper;
	genAmountType genAmount; 
	SF2 sf2;

	sfGenList(SF2 sf2) {
		this.sf2 = sf2;
		sfGenOper = SFGenerator._null;
		genAmount = new genAmountType((short)0);
	}
	//Straightforward constructor
	sfGenList(SF2 sf2, SFGenerator operation, genAmountType amount) {
		this.sf2 = sf2;
		sfGenOper = operation;
		genAmount = amount;
	}

	void write() throws IOException {
		sf2.out.writeShort(Short.reverseBytes(sfGenOper.value));
		sf2.out.writeShort(Short.reverseBytes(genAmount.value));
	}
}

//Instrument zone class
class sfInst {
	String achInstName;
	short wInstBagNdx;
	SF2 sf2;

	//Constructor that automatically points at the end of the (current) preset bag
	sfInst(SF2 sf2, String name) {
		this.sf2 = sf2;
		achInstName = name;
		wInstBagNdx = (short)sf2.pdtaListChunk.ibag_subchunk.bag_list.size();
	}

	void write() throws IOException {
		sf2.writeString20(achInstName);
		sf2.out.writeShort(Short.reverseBytes(wInstBagNdx));
	}
}

class sfSample {
	String achSampleName;
	int dwStart;
	int dwEnd;
	int dwStartloop;
	int dwEndloop;
	int dwSampleRate;
	byte byOriginalPitch;
	byte chPitchCorrection;
	short wSampleLink;
	SFSampleLink sfSampleType;
	SF2 sf2;

	sfSample(SF2 sf2, String name, int Start, int End, int StartLoop,
			int EndLoop, int SampleRate, int OriginalPitch, int PitchCorrection) {
		this.sf2 = sf2;
		achSampleName = name;
		dwStart = Start;
		dwEnd = End;
		dwStartloop = StartLoop;
		dwEndloop = EndLoop;
		dwSampleRate = SampleRate;
		byOriginalPitch = (byte)OriginalPitch;
		chPitchCorrection = (byte)PitchCorrection;
		wSampleLink = 0;
		sfSampleType = SFSampleLink.monoSample;
	}

	void write() throws IOException {
		sf2.writeString20(achSampleName);
		sf2.out.writeInt(Integer.reverseBytes(dwStart));
		sf2.out.writeInt(Integer.reverseBytes(dwEnd));
		sf2.out.writeInt(Integer.reverseBytes(dwStartloop));
		sf2.out.writeInt(Integer.reverseBytes(dwEndloop));
		sf2.out.writeInt(Integer.reverseBytes(dwSampleRate));
		sf2.out.writeByte(byOriginalPitch);
		sf2.out.writeByte(chPitchCorrection);
		sf2.out.writeShort(Short.reverseBytes(wSampleLink));
		sf2.out.writeShort(Short.reverseBytes(sfSampleType.value));
	}
}

enum SFTransform {
	_null(0),
	linear(0),
	concave(1),
	convex(2);

	final short value;

	SFTransform(int value) {
		this.value = (short)value;
	}
}

//Modulator's enumeration class
enum SFModulator {
	_null(0),
	none(0),
	noteOnVelocity(1),
	noteOnKey(2),
	polyPressure(10),
	chnPressure(13),
	pitchWheel(14),
	ptchWeelSensivity(16);
	
	final short value;
	
	SFModulator(int value) {
		this.value = (short)value;
	}
}

//Generator's enumeration class
enum SFGenerator {
	_null(0),
	startAddrsOffset(0),
	endAddrsOffset(1),
	startloopAddrsOffset(2),
	endloopAddrsOffset(3),
	startAddrsCoarseOffset(4),
	modLfoToPitch(5),
	vibLfoToPitch(6),
	modEnvToPitch(7),
	initialFilterFc(8),
	initialFilterQ(9),
	modLfoToFilterFc(10),
	modEnvToFilterFc(11),
	endAddrsCoarseOffset(12),
	modLfoToVolume(13),
	chorusEffectsSend(15),
	reverbEffectsSend(16),
	pan(17),
	delayModLFO(21),
	freqModLFO(22),
	delayVibLFO(23),
	freqVibLFO(24),
	delayModEnv(25),
	attackModEnv(26),
	holdModEnv(27),
	decayModEnv(28),
	sustainModEnv(29),
	releaseModEnv(30),
	keynumToModEnvHold(31),
	keynumToModEnvDecay(32),
	delayVolEnv(33),
	attackVolEnv(34),
	holdVolEnv(35),
	decayVolEnv(36),
	sustainVolEnv(37),
	releaseVolEnv(38),
	keynumToVolEnvHold(39),
	keynumToVolEnvDecay(40),
	instrument(41),
	keyRange(43),
	velRange(44),
	startloopAddrsCoarseOffset(45),
	keynum(46),
	velocity(47),
	initialAttenuation(48),
	endloopAddrsCoarseOffset(50),
	coarseTune(51),
	fineTune(52),
	sampleID(53),
	sampleModes(54),
	scaleTuning(56),
	exclusiveClass(57),
	overridingRootKey(58),
	endOper(60);
	
	final short value;
	
	SFGenerator(int value) {
		this.value = (short)value;
	}
}


//Sample link enumeration
enum SFSampleLink {
	monoSample(1),
	rightSample(2),
	leftSample(4),
	linkedSample(8);

	final short value;

	SFSampleLink(int value) {
		this.value = (short)value;
	}
}

