package gbaMus;

import java.util.ArrayList;
import java.io.RandomAccessFile;
import java.io.IOException;

/*
 * Sound Font Instruments class
 * 
 * This program is part of GBA SoundFontRiper (c) 2012 by Bregalad
 * This is free and open source software.
 */

public class SF2Instr {
	ArrayList<Integer> inst_list;
	SF2 sf2;

	SF2Instr(SF2 sf2) {
		this.sf2 = sf2;
		inst_list = new ArrayList<Integer>();
	}

	//Build a SF2 instrument form a GBA sampled instrument
	int buildSampledInstrument(RandomAccessFile inGBA, int inst_address) throws Exception {

		//Do nothing if this instrument already exists !
		if(inst_list.contains(inst_address)) return inst_list.indexOf(inst_address);

		//Read instrument data
		inGBA.seek(inst_address);

		int instrType = inGBA.read();		//Instrument type
     /* int keynum = */ inGBA.read();			//Key (every key split instrument only)
     /* int unused_byte = */ inGBA.read();		//Unknown/unused byte
     /* int panning = */ inGBA.read();			//Panning (every key split instrument only)

		//The flag is set if no scaling should be done on the sample
		boolean noScale = instrType==8;

		//Get sample pointer
		int sample_pointer = GBAMusRiper.getPointer(inGBA.readInt());

		//Get ADSR envelope
		int adsr = inGBA.readInt();

		//Determine if loop is enabled (it's dumb but we have to seek just for this)
		inGBA.seek(sample_pointer);
		boolean loopFlag = (inGBA.readInt() == 0x40);
		//Build pointed sample
		int SampleIndex = sf2.samples.buildSample(inGBA, sample_pointer);

		//Instrument's name
		String name = "0x" + Integer.toHexString(inst_address) + " sampled";

		//Create instrument bag
		sf2.add_new_instrument(name);
		sf2.add_new_inst_bag();

		//Add generator to prevent scaling if required
		if(noScale)
			sf2.add_new_inst_generator(SFGenerator.scaleTuning, 0);

		generate_adsr_generators(adsr);
		sf2.add_new_inst_generator(SFGenerator.sampleModes, loopFlag ? 1 : 0);
		sf2.add_new_inst_generator(SFGenerator.sampleID, SampleIndex);

		//Add instrument to list
		inst_list.add(inst_address);
		return inst_list.size() - 1;
	}

	//Create new SF2 from every key split GBA instrument
	int buildEveryKeySplitInstrument(RandomAccessFile inGBA, RandomAccessFile PSGData, int Pointer) throws IOException {
		//Do nothing if this instrument already exists !
		if(inst_list.contains(Pointer)) return inst_list.indexOf(Pointer);

		//Add instrument to list
		inst_list.add(Pointer);

		//I'm sorry for doing a dumb copy/pase of the routine right above
		//But there was too much differences to handles to practically handle it with flags
		//therefore I didn't really had a choice.

		sf2.add_new_instrument("0x" + Integer.toHexString(Pointer) + " EveryKeySplit");

		//Get base address
		inGBA.seek(Pointer+4);
		int Baseaddress = GBAMusRiper.getPointer(inGBA.readInt());

		//Loop through all keys
		for(int key = 0; key < 128; key ++) {
			try{
				//Seek at the key's instrument
				inGBA.seek(Baseaddress + 12 * key);

				//Read instrument data
				int instrType = inGBA.read();		//Instrument type
				int keynum = inGBA.read();			//Key (every key split instrument only)
				int unused_byte = inGBA.read();		//Unknown/unused byte
				int panning = inGBA.read();			//Panning (every key split instrument only)

				//The flag is set if no scaling should be done on the sample
				boolean noScale = false;

				int main_word = inGBA.readInt();
				//Get ADSR envelope
				int adsr = inGBA.readInt();

				int SampleIndex;
				boolean loopFlag = true;

				switch(instrType) {
				case 8:
					noScale = true;
				case 0: {
					//Get sample pointer
					int sample_pointer = GBAMusRiper.getPointer(main_word);
					//Determine if loop is enabled (it's dumb but we have to seek just for this)
					inGBA.seek(sample_pointer);
					loopFlag = (inGBA.readInt() == 0x40);
					int pitch = Integer.reverseBytes(inGBA.readInt());
	
					//Build pointed sample
					SampleIndex = sf2.samples.buildSample(inGBA, sample_pointer);
	
					//Add a bag for this key
					sf2.add_new_inst_bag();
					generate_adsr_generators(adsr);
					//Add generator to prevent scaling if required
					if(noScale)
						sf2.add_new_inst_generator(SFGenerator.scaleTuning, 0);
	
					// Compute base note and fine tune from pitch
					double delta_note = 12
							* Math.log((double) sf2.DefaultSampleRate * 1024 / pitch)
							/ Math.log(2);
					double int_delta_note = Math.rint(delta_note);
					int RootKey = 60 + (int)int_delta_note;
	
					//Override root key with the value we need
					sf2.add_new_inst_generator(SFGenerator.overridingRootKey, RootKey - keynum + key);
					//Key range is only a single key (obviously)
					sf2.add_new_inst_generator(SFGenerator.keyRange, key, key);
				}break;

				case 4:
				case 12: {
					//Determine whenever the note is metallic noise, normal noise, or invalid
					boolean metal_flag;
					if(main_word == 0x1000000)
						metal_flag = true;
					else if(main_word == 0)
						metal_flag = false;
					else throw new Exception();

					//Build corresponding sample
					SampleIndex = sf2.samples.buildNoiseSample(PSGData, metal_flag, keynum);
					sf2.add_new_inst_bag();
					generate_psg_adsr_generators(adsr);
					sf2.add_new_inst_generator(SFGenerator.overridingRootKey, key);
					sf2.add_new_inst_generator(SFGenerator.keyRange, key, key);
				}break;
				
				default : throw new Exception();
				}

				if(panning != 0)
					sf2.add_new_inst_generator(SFGenerator.pan, (int)((panning-192) * (500/128.0)));
				//Same as a normal sample
				sf2.add_new_inst_generator(SFGenerator.sampleModes, loopFlag ? (short)1 : 0);
				sf2.add_new_inst_generator(SFGenerator.sampleID, SampleIndex);

			}catch(Exception e) {}
		}

		return inst_list.size() - 1;
	}

	//Build a SF2 instrument from a GBA key split instrument
	int buildKeySplitInstrument(RandomAccessFile inGBA, int Pointer) throws IOException {
		//Do nothing if this instrument already exists !
		if(inst_list.contains(Pointer)) return inst_list.indexOf(Pointer);


		//Get instrument and key table
		inGBA.seek(Pointer+4);
		int BasePointer = GBAMusRiper.getPointer(inGBA.readInt());
		int KeyTable = GBAMusRiper.getPointer(inGBA.readInt());

		//Decode key-table in usable data
		ArrayList<Byte> splitList = new ArrayList<Byte>();
		ArrayList<Integer> indexList = new ArrayList<Integer>();
		byte Key = 0;
		int prev_index = -1;
		int current_index;
		inGBA.seek(KeyTable);

		//Add instrument to list
		inst_list.add(Pointer);
		sf2.add_new_instrument("0x" + Integer.toHexString(Pointer) + " key split");

		do{
			int index = inGBA.readByte();

			//Detect where there is changes in the index table
			current_index = index;
			if(prev_index != current_index) {
				splitList.add(Key);
				indexList.add(current_index);
				prev_index = current_index;
			}
		}while(++Key > 0);
		//Final entry for the last split
		splitList.add((byte)128);

		for(int i=0; i<indexList.size(); i++) {
			try{
				//Seek to pointed instrument
				inGBA.seek(BasePointer + 12 * indexList.get(i));
	
				//Once again I'm sorry for the dumb copy/pase
				//but doing it all with flags would have been quite complex
	
				int instrType = inGBA.read();		//Instrument type
		     /* int keynum = */ inGBA.read();		//Key (every key split instrument only)
		     /* int unused_byte = */ inGBA.read();	//Unknown/unused byte
		     /* int panning = */ inGBA.read();		//Panning (every key split instrument only)
	
				//The flag is set if no scaling should be done on the sample
				boolean noScale = instrType==8;
	
				//Get sample pointer
				int sample_pointer = GBAMusRiper.getPointer(inGBA.readInt());

				//Get ADSR envelope
				int adsr = inGBA.readInt();
	
				//For now GameBoy instruments aren't supported
				//(I wonder if any game ever used this)
				if(instrType!=0 && instrType !=8) continue;
	
				//Determine if loop is enabled (it's dumb but we have to seek just for this)
				inGBA.seek(sample_pointer);
				boolean loopFlag = (inGBA.readInt() == 0x40);
				//Build pointed sample
				int SampleIndex = sf2.samples.buildSample(inGBA, sample_pointer);
	
				//Create instrument bag
				sf2.add_new_inst_bag();
	
				//Add generator to prevent scaling if required
				if(noScale)
					sf2.add_new_inst_generator(SFGenerator.scaleTuning, 0);
	
				generate_adsr_generators(adsr);
				//Particularity here : An additional bag to select the key range
				sf2.add_new_inst_generator(SFGenerator.keyRange, splitList.get(i), splitList.get(i+1)-1);
				sf2.add_new_inst_generator(SFGenerator.sampleModes, loopFlag ? 1 : 0);
				sf2.add_new_inst_generator(SFGenerator.sampleID, SampleIndex);
			}catch(Exception e) {}
		}
		return inst_list.size() - 1;
	}

	void generate_adsr_generators(int adsr) {
		//Get separate components
		int attack = adsr>>>24;
		int decay = (adsr>>>16) & 0xFF;
		int sustain = (adsr>>>8) & 0xFF;
		int release = adsr & 0xFF;

		//Add generators for ADSR envelope if required
		if(attack != 0xFF) {
			//Compute attack time - the sound engine is called 60 times per second
			//and adds "attack" to envelope every time the engine is called
			double att_time = (256/60.0) / attack;
			double att = 1200 * log2(att_time);
			sf2.add_new_inst_generator(SFGenerator.attackVolEnv, (short)att);
		}

		if(sustain != 0xFF) {
			double sus;
			//Compute attenuation in cB if sustain is non-zero
			if(sustain != 0) sus = 100 * Math.log(256.0/sustain);
			//Special case where attenuation is infinite -> use max value
			else sus = 1000;

			sf2.add_new_inst_generator(SFGenerator.sustainVolEnv, (short)sus);

			double dec_time = (Math.log(256.0) /(Math.log(256)-Math.log(decay)))/60.0;
			dec_time *= 10/Math.log(256);
			double dec = 1200 * log2(dec_time);
			sf2.add_new_inst_generator(SFGenerator.decayVolEnv, (short)dec);
		}

		if(release != 0x00) {
			double rel_time = (Math.log(256.0)/(Math.log(256)-Math.log(release)))/60.0;
			double rel = 1200 * log2(rel_time);
			sf2.add_new_inst_generator(SFGenerator.releaseVolEnv, (short)rel);
		}
	}

	//Build gameboy channel 3 instrument
	int buildGB3Instrument(RandomAccessFile inGBA, int inst_address) throws IOException {
		//Do nothing if this instrument already exists !
		if(inst_list.contains(inst_address)) return inst_list.indexOf(inst_address);

		//Read instrument data
		inGBA.seek(inst_address + 4);

		//Get sample pointer
		int sample_pointer = GBAMusRiper.getPointer(inGBA.readInt());
		if (sample_pointer<0 || sample_pointer >= inGBA.length())
			throw new IOException();

		//Get ADSR envelope
		int adsr = inGBA.readInt();

		int sample = sf2.samples.buildGB3Samples(inGBA, sample_pointer);
		String name = "@0x" + inst_address + " GB3";

		sf2.add_new_instrument(name);
		//Global zone
		sf2.add_new_inst_bag();
		generate_psg_adsr_generators(adsr);

		sf2.add_new_inst_bag();
		sf2.add_new_inst_generator(SFGenerator.keyRange, 0, 64);
		sf2.add_new_inst_generator(SFGenerator.sampleModes, 1);
		sf2.add_new_inst_generator(SFGenerator.sampleID, sample);
		sf2.add_new_inst_bag();
		sf2.add_new_inst_generator(SFGenerator.keyRange, 65, 76);
		sf2.add_new_inst_generator(SFGenerator.sampleModes, 1);
		sf2.add_new_inst_generator(SFGenerator.sampleID, sample+1);
		sf2.add_new_inst_bag();
		sf2.add_new_inst_generator(SFGenerator.keyRange, 77, 127);
		sf2.add_new_inst_generator(SFGenerator.sampleModes, 1);
		sf2.add_new_inst_generator(SFGenerator.sampleID, sample+2);

		inst_list.add(inst_address);
		return inst_list.size() - 1;
	}

	//Build GameBoy pulse wave instrument
	int buildPulseInstrument(RandomAccessFile inGBA, RandomAccessFile PSGData, int inst_address) throws IOException {
		if(inst_list.contains(inst_address)) return inst_list.indexOf(inst_address);

		inGBA.seek(inst_address + 4);
		int DutyCycle = inGBA.readInt();

		if((DutyCycle & 0xFFFFFF) != 0) throw new IOException();

		DutyCycle >>>= 24;

		//The difference between 75% and 25% duty cycles is inaudible therefore
		//I simply replace 75% duty cycles by 25%
		if(DutyCycle == 3) DutyCycle = 1;

		if(DutyCycle > 3) throw new IOException();

		int adsr = inGBA.readInt();

		int sample = sf2.samples.buildPulseSamples(PSGData, DutyCycle);
		String name = "0x" + inst_address + " GB pulse";

		sf2.add_new_instrument(name);

		//Global zone
		sf2.add_new_inst_bag();
		generate_psg_adsr_generators(adsr);

		sf2.add_new_inst_bag();
		sf2.add_new_inst_generator(SFGenerator.keyRange, 0, 45);
		sf2.add_new_inst_generator(SFGenerator.sampleModes, 1);
		sf2.add_new_inst_generator(SFGenerator.sampleID, sample);
		sf2.add_new_inst_bag();
		sf2.add_new_inst_generator(SFGenerator.keyRange, 46, 57);
		sf2.add_new_inst_generator(SFGenerator.sampleModes, 1);
		sf2.add_new_inst_generator(SFGenerator.sampleID, sample+1);
		sf2.add_new_inst_bag();
		sf2.add_new_inst_generator(SFGenerator.keyRange, 58, 69);
		sf2.add_new_inst_generator(SFGenerator.sampleModes, 1);
		sf2.add_new_inst_generator(SFGenerator.sampleID, sample+2);
		sf2.add_new_inst_bag();
		sf2.add_new_inst_generator(SFGenerator.keyRange, 70, 81);
		sf2.add_new_inst_generator(SFGenerator.sampleModes, 1);
		sf2.add_new_inst_generator(SFGenerator.sampleID, sample+3);
		sf2.add_new_inst_bag();
		sf2.add_new_inst_generator(SFGenerator.keyRange, 82, 127);
		sf2.add_new_inst_generator(SFGenerator.sampleModes, 1);
		sf2.add_new_inst_generator(SFGenerator.sampleID, sample+4);

		inst_list.add(inst_address);
		return inst_list.size() - 1;
	}

	//Build GameBoy white noise instrument
	int buildNoiseInstrument(RandomAccessFile inGBA, RandomAccessFile PSGData, int inst_address) throws IOException {
		if(inst_list.contains(inst_address)) return inst_list.indexOf(inst_address);

		inGBA.seek(inst_address + 4);
		int Metallic = inGBA.readInt();

		if((Metallic & 0xFFFFFF) != 0) throw new IOException();

		//Switch between normal and metallic noise
		Metallic >>>= 24;

		if(Metallic > 1) throw new IOException();
		boolean metallic = Metallic != 0;

		int adsr = inGBA.readInt();

		String name = "0x" + inst_address + " noise";

		sf2.add_new_instrument(name);

		//Global zone
		sf2.add_new_inst_bag();
		generate_psg_adsr_generators(adsr);

		sf2.add_new_inst_bag();
		int sample42 = sf2.samples.buildNoiseSample(PSGData, metallic, 42);
		sf2.add_new_inst_generator(SFGenerator.keyRange, 0, 42);
		sf2.add_new_inst_generator(SFGenerator.sampleModes, 1);
		sf2.add_new_inst_generator(SFGenerator.sampleID, sample42);

		for(int key = 43; key <=77; key++) {
			sf2.add_new_inst_bag();
			int sample = sf2.samples.buildNoiseSample(PSGData, metallic, key);
			sf2.add_new_inst_generator(SFGenerator.keyRange, key, key);
			sf2.add_new_inst_generator(SFGenerator.sampleModes, 1);
			sf2.add_new_inst_generator(SFGenerator.sampleID, sample);		
		}

		sf2.add_new_inst_bag();
		int sample78 = sf2.samples.buildNoiseSample(PSGData, metallic, 78);
		sf2.add_new_inst_generator(SFGenerator.keyRange, 78, 127);
		sf2.add_new_inst_generator(SFGenerator.sampleModes, 1);
		sf2.add_new_inst_generator(SFGenerator.scaleTuning, 0);
		sf2.add_new_inst_generator(SFGenerator.sampleID, sample78);

		inst_list.add(inst_address);
		return inst_list.size() - 1;
	}

	void generate_psg_adsr_generators(int adsr) throws IOException {
		//Get separate components
		int attack = adsr>>>24;
		int decay = (adsr>>>16) & 0xFF;
		int sustain = (adsr>>>8) & 0xFF;
		int release = adsr & 0xFF;

		//Reject instrument if invalid values !
		if(attack > 15 || decay > 15 || sustain > 15 || release > 15) throw new IOException();

		//Add generators for ADSR envelope if required
		if(attack != 0) {
			//Compute attack time - the sound engine is called 60 times per second
			//and adds "attack" to envelope every time the engine is called
			double att_time = attack/5.0;
			double att = 1200 * log2(att_time);
			sf2.add_new_inst_generator(SFGenerator.attackVolEnv, (short)att);
		}

		if(sustain != 15) {
			double sus;
			//Compute attenuation in cB if sustain is non-zero
			if(sustain != 0) sus = 100 * Math.log(15.0/sustain);
			//Special case where attenuation is infinite -> use max value
			else sus = 1000;

			sf2.add_new_inst_generator(SFGenerator.sustainVolEnv, (short)sus);

			double dec_time = decay/5.0;
			double dec = 1200 * log2(dec_time+1);
			sf2.add_new_inst_generator(SFGenerator.decayVolEnv, (short)dec);
		}

		if(release != 0) {
			double rel_time = release/5.0;
			double rel = 1200 * log2(rel_time);
			sf2.add_new_inst_generator(SFGenerator.releaseVolEnv, (short)rel);
		}
	}

	//Base 2 logarithm function
	double log2(double x) {
		return Math.log(x)/Math.log(2.0);
	}
}