package gbaMus;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;

/*
 * Sound Font Samples class
 * 
 * This program is part of GBA SoundFontRiper (c) 2012 by Bregalad
 * This is free and open source software.
 */

public class SF2Samples {
	ArrayList<Integer> samples_list;
	SF2 sf2;

	SF2Samples(SF2 sf2) {
		this.sf2 = sf2;
		samples_list = new ArrayList<Integer>();
	}

	public int buildSample(RandomAccessFile inGBA, int Pointer) throws Exception {
		// Do nothing if sample already exists
		if (samples_list.contains(Pointer))
			return samples_list.indexOf(Pointer);

		// Read sample data
		inGBA.seek(Pointer);

		int loop = inGBA.readInt();
		int pitch = Integer.reverseBytes(inGBA.readInt());
		int loopPos = Integer.reverseBytes(inGBA.readInt());
		int len = Integer.reverseBytes(inGBA.readInt());

		//Now we should make sure the data is coherent, and reject
		//the samples if errors are suspected

		//Detect invalid samples
		boolean loopFlag;
		if(loop == 0x40)
			loopFlag = true;
		else if (loop == 0x00)
			loopFlag = false;
		else {
			//System.out.println("Error : Illegal loop mode detected");
			throw new Exception();
		}

		//Prevent samples which are way too long or too short
		if (len < 16 || len > 0x3FFFFF) {
			//System.out.println("Error : Extremely long sample detected");
			throw new Exception();
		}

		//Prevent ridiculously high or negative pitches
		if(pitch < 0 || pitch >= 0x10000000) {
			//System.out.println("Error : out of range pitch detected");
			throw new Exception();
		}

		//Prevent samples with illegal loop point from happening
		if (loopPos > len-8) {
			System.out.println("Warning : Illegal loop point detected");
			loopPos = 0;
		}

		// Compute SF2 base note and fine tune from GBA pitch
		// GBA pitch is 1024 * Mid_C frequency
		double delta_note = 12
				* Math.log((double) sf2.DefaultSampleRate * 1024 / pitch)
				/ Math.log(2);
		double int_delta_note = Math.rint(delta_note);
		int PitchCorrection = (int)((int_delta_note - delta_note) * 100);
		int OriginalPitch = 60 + (int)int_delta_note;

		//Create (poetic) instrument name
		String name = "Sample @0x" + Integer.toHexString(Pointer);

		// Add the sample to output
		sf2.sdtaListChunk.smpl_subchunk.addNewSample(inGBA, SampleType.signed_8, name,
				Pointer + 16, len, loopFlag, loopPos, OriginalPitch, PitchCorrection);

		samples_list.add(Pointer);

		// Index of the sample
		return samples_list.size() - 1;
	}

	//Build game boy channel 3 sample
	public int buildGB3Samples(RandomAccessFile inGBA, int Pointer) throws IOException {
		// Do nothing if sample already exists
		if (samples_list.contains(Pointer))
			return samples_list.indexOf(Pointer);

		String name = "GB3 @0x" + Integer.toHexString(Pointer);

		sf2.sdtaListChunk.smpl_subchunk.addNewSample(inGBA, SampleType.gameboy_ch3, name + "A",
				Pointer, 128, true, 0, 65, 24, 22050);
		sf2.sdtaListChunk.smpl_subchunk.addNewSample(inGBA, SampleType.gameboy_ch3, name + "B",
				Pointer, 64, true, 0, 77, 24, 22050);
		sf2.sdtaListChunk.smpl_subchunk.addNewSample(inGBA, SampleType.gameboy_ch3, name + "C",
				Pointer, 32, true, 0, 89, 24, 22050);

		//We have to to add multiple entries to have the size of the list in sync
		//with the numeric indexes of samples....
		samples_list.add(Pointer);
		samples_list.add(Pointer);
		samples_list.add(Pointer);

		// Index of the fist sample
		return samples_list.size() - 3;
	}

	//Build square wave sample
	public int buildPulseSamples(RandomAccessFile PSGData, int DutyCycle) throws IOException {
		if(samples_list.contains(DutyCycle))
			return samples_list.indexOf(DutyCycle);

		final String[] DutyCycles = {"12.5%", "25%", "50%"};
		String name = "Square " + DutyCycles[DutyCycle];

		//This data is referenced to my set of recordings
		//stored in "psg_data.raw"
		final int PointerTbl[][] = {
				{0x0000, 0x2166, 0x3c88, 0x4bd2, 0x698a},
				{0x7798, 0x903e, 0xa15e, 0xb12c, 0xbf4a},
				{0xc958, 0xe200, 0xf4ec, 0x10534, 0x11360}
		};

		final int sizeTbl[][] = {
				{0x10b3, 0xd91, 0x7a5, 0xdec, 0x707},
				{0xc53, 0x890, 0x7e7, 0x70f, 0x507},
				{0xc54, 0x976, 0x824, 0x716, 0x36b}
		};

		final int loopSize[] = {
				689, 344, 172, 86, 43
		};

		for(int i = 0; i < 5; i++) {
			sf2.sdtaListChunk.smpl_subchunk.addNewSample(PSGData, SampleType.signed_16, name + ('A' + i),
					PointerTbl[DutyCycle][i], sizeTbl[DutyCycle][i], true,
					sizeTbl[DutyCycle][i]-loopSize[i], 36 + 12 * i, 38, 44100);
			samples_list.add(DutyCycle);
		}

		//Index of the first sample
		return samples_list.size() - 5;
	}

	//Build white noise sample
	public int buildNoiseSample(RandomAccessFile PSGData, boolean metallic, int key) {
		//prevent out of range keys
		if(key < 42) key = 42;
		if(key > 77) key = 76;

		int num = metallic ? 3 + (key-42) : 80 + (key-42);
		if(samples_list.contains(num))
			return samples_list.indexOf(num);

		String name = "Noise " + (metallic ? "metallic " : "normal ") + key;

		final int[] PointerTbl = { 72246, 160446, 248646, 336846, 425046, 513246, 601446, 689646, 777846,
				866046, 954246, 1042446, 1130646, 1218846, 1307046, 1395246, 1483446, 1571646, 1659846,
				1748046, 1836246, 1924446, 2012646, 2100846, 2189046, 2277246, 2387493, 2475690, 2552863,
				2619011, 2674134, 2718233, 2756819, 2789893, 2817455, 2839504, 2856041, 2867066, 2872578 };

		final int[] NormalLenTbl = { 88200, 88200, 88200, 88200, 88200, 88200, 88200, 88200, 88200, 88200,
				88200, 88200, 88200, 88200, 88200, 88200, 88200, 88200, 88200, 88200, 88200, 88200, 88200,
				88200, 88200, 110247, 88197, 77173, 66148, 55123, 44099, 38586, 33074, 27562, 22049, 16537,
				11025, 5512, 2756 };

		final int[] MetallicLenTbl = { 43755, 38286, 32817, 27347, 21878, 19143, 16408, 13674, 10939, 9572,
				8204, 6837, 5469, 4786, 4102, 3418, 2735,	2393, 2051, 1709, 1367, 1196, 1026,	855, 684,
				598, 513,	427, 342, 299, 256, 214, 171, 150, 128, 107, 85, 64 };

		sf2.sdtaListChunk.smpl_subchunk.addNewSample(PSGData, SampleType.unsigned_8, name,
				PointerTbl[key-42], metallic ? MetallicLenTbl[key-42] : NormalLenTbl[key-42],
				true, 0, key, 0, 44100);

		samples_list.add(num);
		return samples_list.size() - 1;
	}
}
