package gbaMus;
import java.io.RandomAccessFile;
import java.io.IOException;

public class SappyDetector {

	/**
	 * GBA Sappy Engine Detector (c) 2012 by Bregalad
	 * This is free and open source software
	 * 
	 * This java programs detects if a snappy sound engine is present in a given GBA ROM.
	 * If an engine is present it returns a pointer to the instrument list.
	 * If no engine is present, it returns the value (-1)
	 * It's not 100% accurate, so sometimes it might produce erroneous results
	 * 
	 */

	final static String[] Freq = {"invalid", "5734 Hz", "7884 Hz", "10512 Hz", "13379 Hz", "15768 Hz", "18157 Hz",
			"21024 Hz", "26758 Hz", "31536 Hz", "36314 Hz", "40137 Hz", "42048 Hz", "invalid", "invalid", "invalid"};

	public static void main(String[] args) throws IOException {
		if(args.length != 1) printInstructions();

		RandomAccessFile inGBA = new RandomAccessFile(args[0], "r");
		
		sappyDetector(inGBA);

		inGBA.close();
	}

	public static int sappyDetector(RandomAccessFile inGBA) throws IOException {
		int data, dummy, offset = -1;
		boolean re_seek_flag = true;
		boolean detected = false;
		
		System.out.println("Sappy sound engine detector (c) 2012 by Bregalad");
		System.out.print("Searching");

		//This loops try to look for particular opcodes in the whole ROM
		//and this enables the program to detect the sound engine's location

		for(int i=0; i<inGBA.length()-4 && !detected; i+=4) {
			//Re-seek in file only if necessarily (calls to seek() are slow !)
			if(re_seek_flag) {
				inGBA.seek(i);
				re_seek_flag = false;
			}

			//Read 32-bit value from input file
			data = Integer.reverseBytes(inGBA.readInt());	
	
			//Look for a particular opcode sequence...
			if (data == 0x4700bc01 || data == 0x4700) {
				inGBA.seek(i-4);
				re_seek_flag = true;
				dummy = inGBA.readInt();

				if (dummy == 0xf1d170bc || dummy == 0xf0bc01bc) {
					inGBA.seek(i-8);
					dummy = inGBA.readInt();
					if (dummy == 0x13e002e || dummy == 0x08bc9846) {
						inGBA.seek(i+12);
						dummy = inGBA.readInt();
						if(dummy == 0x10004) System.out.println("Sound engine detected : Stereo version");
						else if(dummy == 0xe0000004) System.out.println("Sound engine detected : Mono version");
						else continue;

						//At this point, we PROBABLY detected the sound engine but it might
						//still be a false positive...
						inGBA.seek(i+24);
						dummy = inGBA.readInt();
						
						//Get engine's (supposed ?) main parameters
						int main_vol = (dummy & 0xF00000) >> 20;
						int polyphony = (dummy & 0xF0000)>> 16;
						int dac_bits = 17-((dummy & 0xF000) >> 12);
						int sampling_rate_index = (dummy &0xF00)>>8;

						//Compute (supposed ?) address of song table
						inGBA.seek(i+28);
						dummy = Integer.reverseBytes(inGBA.readInt());	//Read # of song levels
						int SongTbl = (Integer.reverseBytes(inGBA.readInt())&0x3FFFFFF) + 12 * dummy;

						//Prevent illegal values for all fields
						if(main_vol == 0 || polyphony > 12 || dac_bits < 6 || dac_bits > 9 ||
								sampling_rate_index < 1 || sampling_rate_index > 12 || SongTbl >= inGBA.length()) {
							System.out.print("Fake alert. Continue searching");
							continue;
						}

						//At this point we can be almost certain we detected the real thing.
						System.out.print("Engine parameters : ");
						System.out.print("Main Volume : " + main_vol);
						System.out.print(", Polyphony : " + polyphony + " channels");
						System.out.print(", Dac : " + dac_bits + " bits");
						System.out.println(", Sampling rate : " + Freq[sampling_rate_index] );
						System.out.println("Song table located at : 0x" + Integer.toHexString(SongTbl));

						detected = true;
						offset = i;
					}
				}
			}
			//Regularly print a small '.' when searching through the file
			if ((i&0x1FFFF)==0) System.out.print(".");
		}

		if(!detected) {			//If no address were told manually and nothing was detected....
			System.out.println("\nNo sappy engine was found.");
		}
		return(offset);
	}

	static void printInstructions() {
		System.out.println("GBA Sappy Engine Detector (c) 2012 by Bregalad");
		System.out.println("Usage : SappyDetector game.gba");
		System.exit(-1);
	}

}
